/*
see README.txt for details.

chris <cliechti@gmx.net>
*/
#include "hardware.h"
#include <stdlib.h>
#include <stdio.h>
#include "lcd.h"

//the font file has no separate header file, it just provides
//the font in this array:
extern const char lcdfont[64];

//keymap for the AT keyboard SG layout
typedef struct {
    char scancode;
    char unshifted;
    char shifted;
} keytranslation;

#define SHIFTED (1<<0)
#define ALTERED (1<<1)

const keytranslation keymap[] = {
    {0x1c, 'a', 'A'},
    {0x32, 'b', 'B'},
    {0x21, 'c', 'C'},
    {0x23, 'd', 'D'},
    {0x24, 'e', 'E'},
    {0x2b, 'f', 'F'},
    {0x34, 'g', 'G'},
    {0x33, 'h', 'H'},
    {0x43, 'i', 'I'},
    {0x3b, 'j', 'J'},
    {0x42, 'k', 'K'},
    {0x4b, 'l', 'L'},
    {0x3a, 'm', 'M'},
    {0x31, 'n', 'N'},
    {0x44, 'o', 'O'},
    {0x4d, 'p', 'P'},
    {0x15, 'q', 'Q'},
    {0x2d, 'r', 'R'},
    {0x1b, 's', 'S'},
    {0x2c, 't', 'T'},
    {0x3c, 'u', 'U'},
    {0x2a, 'v', 'V'},
    {0x1d, 'w', 'W'},
    {0x22, 'x', 'X'},
    {0x1a, 'y', 'Y'},
    {0x35, 'z', 'Z'},
    {0x16, '1', '+'},
    {0x1e, '2', '"'},
    {0x26, '3', '*'},
    {0x25, '4', ''},
    {0x2e, '5', '%'},
    {0x36, '6', '&'},
    {0x3d, '7', '/'},
    {0x3e, '8', '('},
    {0x46, '9', ')'},
    {0x45, '0', '='},
    {0x29, ' ', ' '},
    {0x41, ',', ';'},
    {0x49, '.', ':'},
    {0x4a, '-', '_'},
    {0x76, '\x1b', '\x1b'},   //ESC
    {0x5a, '\n', '\n'},       //RETURN
    {0x0d, '\t', '\t'},       //TAB
    {0x66, '\b', '\b'},       //BS
    {0}       //sentinel
};

char findkey(char scancode, int mode) {
    const keytranslation *p = keymap;
    while (p->scancode) {
        if (p->scancode == scancode) {
            switch (mode) {
                case SHIFTED:
                    return p->shifted;
                case ALTERED:
                case ALTERED|SHIFTED:
                default:
                    return p->unshifted;
            }
        }
        p++;
    }
    return 0;
}

//redirect stdio looking funcs to LCD
#define printf(format, args...) uprintf(lcdPutc, format, args)
#define puts            lcdPuts
#define putc            lcdPutc

//global vars to store the time
int h, m, s;

/**
This one is executed onece a second. it counts seconds, minues, hours - hey
it shoule be a clock ;-)
it does not count days, but i think you'll get the idea.
*/
wakeup interrupt (TIMERA0_VECTOR) INT_TimerA_CCR0(void) {
    CCR0 += ONESECOND;                  //setup next irq
    if (++s > 59) {                     //count seconds, up to one minute
        s = 0;                          //reset seconds and ...
        if (++m > 59) {                 //count miutes, up to one hour
            m = 0;                      //reset minutes and ...
            if (++h > 23) {             //cound hours, up to one day
                h = 0;                  //reset hours and ...
                //++day ...             //don't do anything, could count days here
            }
        }
    }
}

/**
Delay function.
*/
void delay(unsigned int d) {
   int i;
   for (i = 0; i<d; i++) {
      nop();
      nop();
   }
}


#define CLK_HI  P2DIR &= ~CLK
#define CLK_LO  P2DIR |= CLK
#define DAT_HI  P2DIR &= ~DAT
#define DAT_LO  P2DIR |= DAT
#define WAIT_NEG_PULSE  while ((P2IN & CLK) != 0); while ((P2IN & CLK) == 0)
#define WAIT_NEG  while ((P2IN & CLK) != 0)
#define WAIT_POS  while ((P2IN & CLK) == 0)
#define HIGHLOW   while ((P2IN & CLK) == 0); while ((P2IN & CLK) != 0)

char rawqueue[8];      //store raw scancodes
int  rawq_start;
int  rawq_end;

wakeup interrupt (PORT2_VECTOR) INT_P2(void) {
    unsigned char key = 0;
    int i;
    WAIT_POS;             //wait until startbit is finished
    for (i=0; i<8; i++) {
        WAIT_NEG;
        key >>= 1;
        key |= (P2IN & DAT) ? 0x80 : 0;
        WAIT_POS;
    }
    WAIT_NEG;               //wait for parity
    //p = (P2IN & DAT) ? 0x80 : 0;
    WAIT_POS;
    WAIT_NEG;               //wait for stop bit
    //WAIT_POS;
    rawqueue[rawq_end] = key;
    if (++rawq_end >= sizeof(rawqueue)) rawq_end = 0;
    P2IFG &= ~(CLK|DAT);
}

#define QSIZE 8
int queue[QSIZE];      //store keycodes
int q_start;
int q_end;

#define RELEASE         0x8000
#define SPECIALKEY      0x4000
#define CAPS    (SPECIALKEY|0x58)
#define F1      (SPECIALKEY|0x05)
#define F2      (SPECIALKEY|0x06)
#define F3      (SPECIALKEY|0x04)
#define F4      (SPECIALKEY|0x0c)
#define F5      (SPECIALKEY|0x03)
#define F6      (SPECIALKEY|0x0b)
#define F7      (SPECIALKEY|0x83)
#define F8      (SPECIALKEY|0x0a)
#define F9      (SPECIALKEY|0x01)
#define F10     (SPECIALKEY|0x09)
#define F11     (SPECIALKEY|0x78)
#define F12     (SPECIALKEY|0x07)
void processRawq(void) {
    static int mode = 0;
    static int nextrelease = 0;
    
    while (rawq_start != rawq_end) { //queue not empty
        unsigned char key;
        int flags = 0;
        int c;
        //protect queue by disabling irqs during reading it
        dint();
        key = rawqueue[rawq_start];
        if (++rawq_start >= sizeof(rawqueue)) rawq_start = 0;
        eint();
        //ok, now work on copied value
        switch (key) {
            case 0xf0: //release key
                nextrelease = 1;
                break;
            case 0x12: //left SHIFT
            case 0x59: //right SHIFT
                if (nextrelease) {
                    mode &= ~SHIFTED;
                } else {
                    mode |= SHIFTED;
                }
                nextrelease = 0;
                break;
            case 0x11: //left ALT
                if (nextrelease) {
                    mode &= ~ALTERED;
                } else {
                    mode |= ALTERED;
                }
                nextrelease = 0;
                break;
            default:
                if (nextrelease) flags = RELEASE;
                c = findkey(key, mode);
                if (c) {
                    queue[q_end] = c | flags;
                } else {
                    queue[q_end] = key | flags | SPECIALKEY;
                }
                if (++q_end >= QSIZE) q_end = 0;
                nextrelease = 0;
        }
    }
}

//get a key from the queue
int getkey(void) {
    int c;
    if (q_start == q_end) return 0;     //return 0 if queue empty
    c = queue[q_start];
    if (++q_start >= QSIZE) q_start = 0;
    return c;
}

void sendkeyb(char cmd) {
    int i;
    int p = 0;
    //port as output
    P2IE  &= ~CLK;              //disable irqs while sending
    P2OUT &= ~(CLK|DAT);        //make sure the out reg is 0
    //clk pulse
    CLK_LO;
    delay(10);
    DAT_LO;
    CLK_HI;
    //startbit
    HIGHLOW;
    //data
    for (i = 0; i<8; i++) {
        if (cmd & 1) {
            DAT_HI;
            p++;
        } else {
            DAT_LO;
        }
        HIGHLOW;
        cmd >>= 1;
    }
    //parity
    if (p&1) {
        DAT_LO;
    } else {
        DAT_HI;
    }
    HIGHLOW;
    //stop
    DAT_HI;        //that's it, wait for ACK
    HIGHLOW;               //stop bit from keyb
    //wait for ack
    while ((P2IN & CLK) != 0) ;
    if (P2IN & DAT) {
        //blink once
        P2OUT |= LED;
        delay(1000);
        P2OUT &= ~LED;
    } else {
        //blink twice
        P2OUT |= LED;
        delay(1000);
        P2OUT &= ~LED;
        delay(10000);
        P2OUT |= LED;
        delay(1000);
        P2OUT &= ~LED;
    }
    while ((P2IN & CLK) == 0) ;

    //restore interrupt handler function
    P2IFG &= ~CLK;
    P2IE |= CLK;
}


/**
Main function with ...
*/
int main(void) {
    int i = 0, cursor=0;
    unsigned int c;

    WDTCTL = WDTCTL_INIT;               //Init watchdog timer

    P1OUT  = P1OUT_INIT;                //Init output data of port1
    P2OUT  = P2OUT_INIT;                //Init output data of port2
    P3OUT  = P3OUT_INIT;                //Init output data of port3

    P1SEL  = P1SEL_INIT;                //Select port or module -function on port1
    P2SEL  = P2SEL_INIT;                //Select port or module -function on port2
    P3SEL  = P3SEL_INIT;                //Select port or module -function on port3

    P1DIR  = P1DIR_INIT;                //Init port direction register of port1
    P2DIR  = P2DIR_INIT;                //Init port direction register of port2
    P3DIR  = P3DIR_INIT;                //Init port direction register of port3

    P1IES  = P1IES_INIT;                //init port interrupts
    P2IES  = P2IES_INIT;
    P1IE   = P1IE_INIT;
    P2IE   = P2IE_INIT;

    BCSCTL1 |= DIVA0 | DIVA1;           //setup divider for RTC
    TACTL  = TACTL_INIT;                //setup timer (still stopped)
    CCR0   = ONESECOND;                 //setup first RTC irq
    CCTL0  = CCIE;                      //enable compare interrupts
    CCTL1  = 0;                         //the other modules are not used
    CCTL2  = 0;                         //...
    TACTL  |= MC1;                      //start timer

    eint();                             //enable interrupts

    P2OUT |= LED;                       //light LED during init
    lcdOn();                            //reset sequence
    lcdInit();                          //setup LCD modes
    P2OUT &= ~LED;                      //switch off LED

    sendkeyb(0xed);
    delay(1000);
    sendkeyb(BIT0|BIT1);
    puts("AT keyboard demo");           //say hello

    while (1) {                         //main loop, never ends...
        lcdInstr(LCDLINE2);
        //show the clock
        //as we are synced with the interrupt, its save to read h,m,s
        //otherwise a dint..eint around the following line would be
        //very helpful to prevent bogus values
        printf("%02d:%02d:%02d", h, m, s);
        //now show the bargraph. it hasn't to do much with a clock
        //but we want to show the feature here...
        lcdInstr(LCDLINE2+15);
        putc(i);
        if (++i >= 8) i = 0;            //lets show all 8 downloadable chars
        LPM0;                           //sync, wakeup by irq
        processRawq();
        while( (c = getkey()) ) {
            lcdInstr(LCDLINE2+10);
            printf("%04x", c);
            if ((c >> 8) == 0) {
                lcdInstr(LCDLINE1 + cursor);
                if (++cursor >= 16) cursor = 0;
                putc(c);
            } else {
                switch (c) {
                    case CAPS:
                        sendkeyb(0xed);
                        delay(1000);
                        sendkeyb((s & 1) ? (BIT0|BIT1) : 0);
                        break;
                    case F1:
                        lcdInstr(LCDLINE1);
                        puts("                ");
                        cursor = 0;
                }
            }
        }
        //lcdInstr(LCDLINE2+12);
        //printf("%02x", key);
    }
}

