/******************************************************************************
 *
 * $Workfile: task.c $
 *
 * This module provides the code for a Co-operative Multi-tasker
 * Copyright 1996-1997,2000-2002, R O SoftWare
 *
 * No guarantees, warrantees, or promises, implied or otherwise.
 * May be used for hobby or commercial purposes provided copyright
 * notice remains intact.
 *
 * $History: $
 * 
 *****************************************************************************/
#include <stdlib.h>
#include <io.h>
#include "config.h"
#include "task.h"


typedef struct _TASK
  {
  void         **stack;                 // current stack pointer
  struct _TASK *next;                   // next task in chain
  struct _TASK *prev;                   // previous task in chain
  } TASK;


static TASK  tasks[MAX_TASKS];
static TASK  *currentTask;
static short totalTasks;

static TASK *makeTask(void **stackBase);
static void die(void);


#define saveRegs()  \
  __asm__ __volatile__ ("push R4\n"  \
                        "push R5\n"  \
                        "push R6\n"  \
                        "push R7\n"  \
                        "push R8\n"  \
                        "push R9\n"  \
                        "push R10\n" \
                        "push R11")


#define restoreRegs() \
   __asm__ __volatile__ ("pop R11\n"  \
                         "pop R10\n"  \
                         "pop R9\n"   \
                         "pop R8\n"   \
                         "pop R7\n"   \
                         "pop R6\n"   \
                         "pop R5\n"   \
                         "pop R4")


/******************************************************************************
 *
 * Function Name: makeTask()
 *
 * Description:
 *    This function finds an open task slot and initializes it.
 *
 * Calling Sequence:
 *    void **stackBase - pointer to base of the stack
 *
 * Returns:
 *    ID of created task
 *    NULL on error
 *
 *****************************************************************************/
static TASK *makeTask(void **stackBase)
{
  int i;
  TASK *t;

  // loop through the array of tasks
  // looking for an open slot
  for (i = 0; i < MAX_TASKS; i++)
    {
    if (tasks[i].stack == NULL)
      break;
    }

  // find an open one
  if (i == MAX_TASKS)
    return NULL;                        // no, return NULL

  t = &tasks[i];                        // get pointer to it

  // check if a stack has been provided
  // (required for all 'fork'ed tasks)
  if (stackBase)
    {
    t->stack = stackBase;
    t->prev = currentTask->prev;
    t->next = currentTask;
    t->prev->next = t;
    t->next->prev = t;
    }
  else
    {
    // this must be the call via initTasking()
    // save the current stack pointer
    asm(" mov.w R1,%0" : "=r" (t->stack));
    t->next = t;
    t->prev = t;
    }

  totalTasks++;                         // increment the task count
  return t;
}


/******************************************************************************
 *
 * Function Name: fork()
 *
 * Description:
 *    This function installs the referenced function as a task using
 *    the referenced stack.
 *
 * Calling Sequence:
 *    void (*f) - address of task function
 *    void **stackBase - pointer to base of the stack
 *
 * Returns:
 *    task ID of newly created task
 *    NULL on error
 *
 *****************************************************************************/
void *fork(void (*f)(void), void **stackBase)
{
  TASK *t;

  // don't do anything w/o calling initTasking() first
  // must also provide a stack
  if (!currentTask || !stackBase)
    return NULL;

  // make the task
  if ((t = makeTask(stackBase)) != NULL)
    {
    // push address of die() for when f() returns
    *(--t->stack) = (void *)die;

    // push address of f() so swap() can start it
    *(--t->stack) = (void *)f;

    // make space for dummy R4..R11 for swap() to recover
    t->stack -= 8;
    }

  return t;
}


/******************************************************************************
 *
 * Function Name: initTasking()
 *
 * Description:
 *    This function initializes the tasker, setting up the calling
 *    function as a task using the current SP for the stack.
 *
 * Calling Sequence:
 *    void
 *
 * Returns:
 *    task ID of the current task
 *
 *****************************************************************************/
void *initTasking(void)
{
  // if first time through, set up the world
  if (!currentTask)
    currentTask = makeTask(NULL);

  return currentTask;
}


/******************************************************************************
 *
 * Function Name: swap()
 *
 * Description:
 *    This function saves the current task state, services the watchdog
 *    timer, changes to the next task state, then returns.
 *
 * Calling Sequence:
 *    void
 *
 * Returns:
 *    void
 *
 *****************************************************************************/
void swap(void)
{
  saveRegs();

#ifdef WDOG
  // keep the watchdog timer alive
  WDOG();
#endif

  // make sure we have been initialized
  if (currentTask)
    {
    // save the current stack pointer
    __asm__ __volatile__ ("mov.w R1,%0" : "=r" (currentTask->stack));

    // get the next task pointer from the list
    currentTask = currentTask->next;

    // and recover its stack pointer
    __asm__ __volatile__ ("mov.w %0,R1" : : "m" (currentTask->stack));
    }

  restoreRegs();                        // recover registers and return
}


/******************************************************************************
 *
 * Function Name: getTaskID()
 *
 * Description:
 *    This function returns the current task ID
 *
 * Calling Sequence:
 *    void
 *
 * Returns:
 *    task ID of the current task
 *
 *****************************************************************************/
void *getTaskID(void)
{
  return currentTask;
}


/******************************************************************************
 *
 * Function Name: die()
 *
 * Description:
 *    This function terminates the calling task.
 *
 * Calling Sequence:
 *    void
 *
 * Returns:
 *    void
 *
 *****************************************************************************/
static void die(void)
{
  terminate(NULL);
}


/******************************************************************************
 *
 * Function Name: terminate()
 *
 * Description:
 *    This function deletes the referenced task from the task list,
 *    changes to the next task state, then returns.
 *
 * Calling Sequence:
 *    *taskID - task to terminate, current task if NULL
 *
 * Returns:
 *    void
 *
 *****************************************************************************/
void terminate(void *taskID)
{
  int wasCurrentTask = 0;

  // check for a default request to delete the current task
  if (taskID == NULL)
    taskID = currentTask;               // make it explicit

  // check if it is a request to delete the current task
  if (taskID == currentTask)
    {
    currentTask = ((TASK *)taskID)->next; // set currentTask to something valid
    wasCurrentTask = 1;
    }

  // cut the task out of the task list
  ((TASK *)taskID)->prev->next = ((TASK *)taskID)->next;
  ((TASK *)taskID)->next->prev = ((TASK *)taskID)->prev;

  // mark the task struct unused
  ((TASK *)taskID)->stack = NULL;

  // check for deletion of the last task
  if (!--totalTasks)
    exit();

  // check if it was the current task
  if (!wasCurrentTask)
    return;                             // no, nothing else to do

  // yes, get the new task state
  // and recover its stack pointer
  __asm__ __volatile__ (" mov.w %0,R1" : : "m" (currentTask->stack));
  restoreRegs();                        // recover registers and return
}
