#include "hardware.h"

extern unsigned char videomem[24*4];
extern unsigned int freq;
extern unsigned int sde;
extern unsigned int play;
extern unsigned short line;

#define p_clear(x,y) videomem[(y)*4+(3-((x)>>3))] &= ~(0x80>>((x)&7))
#define p_set(x,y)   videomem[(y)*4+(3-((x)>>3))] |=  (0x80>>((x)&7))
#define p_get(x,y)  (videomem[(y)*4+(3-((x)>>3))] & (0x80>>((x)&7)))

#define is_line(y) ((videomem[y*4+3] == 0xff) && (videomem[y*4+2] == 0xff))

#include "tiles.ci"
#include "numbers.ci"

#define W 32
#define H 24

volatile int px;        //current tile horizontal pos
volatile int py;        //current tile vertical pos
volatile int sticking;  //current tile can move (1 otherwise)
int score;              //number of squashed lines
int level;              //level from 9 (easiest) to 1 (fastest)
char part[5];           //current tile
char nextpart[5];       //buffer for next tile
char newpart[5];        //temp buffer for rotations


void bing( unsigned int fr ) {
    sde = freq = fr;
    play = 100;
}

//pseudo random generator
//low multipicator, only suitable with a modulo of samall number
unsigned char random(void) {
    static char seed = 0;
    seed = (seed*17)+1;
    return seed;
}

//peek at screen coordinate
//return 0 if out of bounds except for the bottom
//where it is a 1
int get(int x, int y) {
    if ((x < 0) || (x >= W) || (y < 0)) {
        return 0;
    } else if (y >= H) {
        return 1;
    } else {
        return p_get(x,y);
    }
}

//scroll playing are starting from y one down
//(collapse lines)
void scroll(int y) {
    int dx,dy;
    for (dy=y; dy>=0; dy--) {
        videomem[4*dy+3] = videomem[4*(dy-1)+3];
        videomem[4*dy+2] = videomem[4*(dy-1)+2];
    }
    videomem[0] = videomem[1] = 0;
}

//copy 5 bytes from src->dst
void copy(const char *src, char *dst) {
    int y;
    for (y=0; y<5; y++) {
        dst[y] = src[y];
    }
}

//rotate part 90deg to the left and save it to newpart
void rotate_left() {
    char save[5];
    int x,y;
    for (y=0; y<5; y++) {
        save[y] = part[y];
        newpart[y] = 0;
    }

    for (x=0; x<5; x++) {
        for (y=0; y<5; y++) {
            if (part[y] & (0x80>>(4-x))) {
                newpart[x] |= 0x80>>y;
            }
        }
    }
}

//rotate part 90deg to the right and save it to newpart
void rotate_right() {
    char save[5];
    int x,y;
    for (y=0; y<5; y++) {
        save[y] = part[y];
        newpart[y] = 0;
    }

    for (x=0; x<5; x++) {
        for (y=0; y<5; y++) {
            if (part[4-y] & (0x80>>(x))) {
                newpart[x] |= 0x80>>y;
            }
        }
    }
}

//can the xpart be placed without conflicts?
//returns true if it fits
int can_place(int x, int y, const char *xpart) {
    int dx,dy;
    for (dy=0; dy<5; dy++) {
        for (dx=0; dx<5; dx++) {
            if (get(x+dx,y+dy) && (xpart[dy] & (0x80>>dx))) {
                return 0;
            }
        }
    }
    return 1;
}

//draw border for playing area
void border(void) {
    int y;
    for (y=0; y<H; y++) {
        videomem[4*y+3] |= 0x80;
        videomem[4*y+2] |= 0x01;
    }
}

//copy symboc "ch" to screen coordinates x,y
//draw==1 => draw | draw==0 => erase
void blit(int x, int y, const char *ch, int draw) {
    int dx,dy;
    for (dy=0; dy<5; dy++) {
        for (dx=0; dx<5; dx++) {
            if (ch[dy] & (0x80 >> dx)) {
                if (draw) {
                    p_set(x+dx, y+dy);
                } else {
                    p_clear(x+dx, y+dy);
                }
            }
        }
    }
}


//show game info on the right side
void info() {
    int x,y;
    int hun = 0;
    int s = score;
    //clear information half of screen
    for (y=0; y<16; y++) {
        videomem[4*y+0] = videomem[4*y+1] = 0;
    }
    //next
    blit(18,6, nextpart,1);
    //draw score
    if (s/100) {
        blit(17,0,numbers[s/100],1);
        s %= 100;
        hun = 1;
    }
    if (s/10 || hun) {
        blit(22,0,numbers[s/10],1);
    }
    blit(27,0,numbers[s%10],1);
    //level
    blit(27,6,numbers[10-level],1);
}

//move current part one line down
void move_down(void) {
    if (can_place(px,py+1,part)) {
        py += 1;
    } else {
        sticking = 1;
    }
}

//games' main loop
void game( void ) {
    int div = 0, my;
    char s;

    //copy some random starting tiles
    copy(tiles[random()%NUMTILES], part);
    copy(tiles[random()%NUMTILES], nextpart);

    //clear screen
    for (my=0; my<H*4; my++) {
        videomem[my] = 0;
    }

    score = 0;
    level = 9;
    px = 6;
    py = 0;
    sticking = 0;
    while (1) {
        P2OUT = 0;      //LED off
        LPM0;   //vsync, slow down game play
        LPM0;   //vsync
        LPM0;   //vsync
        LPM0;   //vsync
        P2OUT = 0xff;   //LED on
        //get input
        s = ~P1IN;
        //videomem[23*4] = s;               //debug
        
        if (sticking) {
            //next becomes this, reset coords
            copy(nextpart, part);
            px = 6;
            py = 0;
            //choose new tile for next one
            copy(tiles[random()%NUMTILES], nextpart);
            sticking = 0;
            //squash lines
            for (my=0; my<H; my++) {
                if (is_line(my)) {
                    scroll(my);
                    score += 1;
                }
            }
            //calculate level from score
            if ((score < 100)) {
                level = 9 - (score/10);
            } else {
                level = 1;
            }
            //is the new tile placeable?
            if (!can_place(px,py,part)) {
                return; //"game over"
            }
        } else {
            blit(px,py,part,0); //hide tile
            //handle inputs
            if (s & RIGHT1) {
                if ((px < W-1) && can_place(px+1,py,part)) {
                    px = px + 1;
                }
            }
            if (s & LEFT1) {
                if ((px >= 0) && can_place(px-1,py,part)) {
                    px = px - 1;
                }
            }
            if (s & UP1) {
                rotate_left();
                if (can_place(px,py,newpart)) {
                    copy(newpart, part);
                }
            }
            if (s & DOWN1) {
                rotate_right();
                if (can_place(px,py,newpart)) {
                    copy(newpart, part);
                }
            }
            if (s & FIRE1) {
                while (!sticking) {
                    move_down();
                }
            }
        }
        //the current tile looses height over time
        if (div++ % level == 0) {//speed depends on level
            move_down();
        }
        blit(px,py,part,1);     //show current part
        info();                 //draw game info
        border();               //draw border around playfield
    }

}


